# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
from Game import GameState

def get_threat_lines(grid, pnum, minlength):
    # only look in "positive" directions
    zeros = (0,)*grid.ndim
    dirs = [d for d in GridUtils.direction_array(grid.ndim) if d>zeros]
    for c in GridUtils.enumerate_coords_with_value(grid, pnum):
        for d in dirs:
            # check back
            goodback = True
            backopen = 0
            for bc,val in GridUtils.enumerate_values_with_direction(grid, c, GridUtils.negate_tuple(d)):
                if val==0:
                    backopen += 1
                elif val==pnum:
                    if backopen==0:
                        goodback = False
                        break # we've already done this line
                    else:
                        backopen += 1
                else:
                    break # non-friendly piece
            if goodback:
                threatlen = 1
                fwdopen = 0
                inthreat = True
                for fc,val in GridUtils.enumerate_values_with_direction(grid, c, d):
                    if val==0:
                        if threatlen==1:
                            break # no line
                        inthreat = False
                        fwdopen += 1
                    elif val==pnum:
                        if inthreat:
                            threatlen += 1
                        else:
                            fwdopen += 1
                    else:
                        break # non-friendly piece
                if threatlen > 1 and backopen+threatlen+fwdopen >= minlength:
                    yield ({'position': c, 'direction': d, 'length': threatlen, 
                                    'open_before': backopen, 'open_after': fwdopen,})

def threats_with_length(grid, val, minlength):
    # yields only full lines minlength or longer
    return (t for t in get_threat_lines(grid, val, minlength) if t['length']>=minlength)
    
def positions_for_threat_line(threat):
    p = threat['position']
    d = threat['direction']
    for i in xrange(threat['length']):
        yield p
        p = GridUtils.add_tuples(p, d)

def threat_weight(gstate, pnum, weights):
    total = 0
    for t in get_threat_lines(gstate.grid, pnum, gstate.win_length()):
        tl = t['length']
        bothopen = t['open_before'] and t['open_after']
        index = 2*tl + (1 if bothopen else 0)
        total += weights[index]
    return total
    
def has_threat_size(tsize, grid, pnum):
    return any(threats_with_length(grid, pnum, tsize))
    
def lines_through_position(grid, pos, minlength):
    """ Yields positions of complete lines through pos with length of at least minlength """
    zeros = (0,)*grid.ndim
    dirs = [d for d in GridUtils.direction_array(grid.ndim) if d>zeros]
    val = grid[pos]
    
    def _process_positions(pgen):
        for linepos,lineval in pgen:
            if lineval==val:
                yield linepos
            else:
                break
                
    for d in dirs:
        # check forward and backward, accumulating positions
        positions = [pos]
        positions.extend(_process_positions(GridUtils.enumerate_values_with_direction(grid, pos, d)))
        positions.extend(_process_positions(GridUtils.enumerate_values_with_direction(grid, pos, GridUtils.negate_tuple(d))))
        if len(positions)>=minlength:
            yield positions
    

class LineGameState(GameState):
    def win_length(self):
        raise NotImplementedError
        
    def has_scores(self):
        return False
            
    def is_game_over(self):
        if 0 not in self.grid: return True
        return has_threat_size(self.win_length(), self.grid, self.previous_player())
        
    def winning_player(self):
        if has_threat_size(self.win_length(), self.grid, self.previous_player()):
            return self.previous_player()
        return 0
    
def test():
    import numpy
    a = numpy.zeros((8,8), int)
    t1 = list(get_threat_lines(a, 1, 5))
    
    a[3,3] = a[4,4] = a[0,1] = a[0,0] = 1
    t2 = list(get_threat_lines(a, 1, 5))
    
    a[1,1] = 2
    t3 = list(get_threat_lines(a, 1, 5))

    a[5,5] = 2
    t4 = list(get_threat_lines(a, 1, 5))

    return (a,t1,t2,t3,t4)
    